#include <config.h>

#include <string.h>
#include <stdio.h>
#include "py_module.h"

#include "data.h"


void data_read_rest_of_line(FILE *stream, gchar *str, gint max_length)
{
     gchar buf[1024];
     gint length = 0;

     fgets(buf, 1024, stream);
     length = strlen(buf);
     if (buf[strlen(buf)-1] == '\n')
	  buf[strlen(buf)-1] = '\0';
     g_strstrip(buf); /* removes leading and trailing whitespace */

     str[max_length-1] = '\0';
     strncpy(str, buf, (size_t)max_length-1);
}

void data_read(const gchar *filename)
{
    FILE *stream;
    AccountData **acc_table;	/* temporary table to map account's id onto the pointer to this account's data */

    if (!strcmp(filename, "-"))
	 stream = stdin;
    else
	stream = fopen(filename, "r");

    if (stream == NULL) {
	 g_printerr("Failed to open %s\n", filename);
	 return;
    }

    /* read accounts data */
    {
	int acc_number, i;

	if (fscanf(stream, "%d\n", &acc_number) == EOF)
	     return;

	acc_table = g_new(AccountData *, acc_number);

	/* allocate 'acc_number' AccountData structures and link them into a list */
	for (i = 0; i < acc_number; i++) {
	     acc_table[i] = g_new(AccountData, 1);
	     acc_list = g_list_prepend(acc_list, acc_table[i]);
	}

	/* and now read those accounts into the table */
	for (i = 0; i < acc_number; i++) {
	     int account_id, parent_id;

	     fscanf(stream, "%d %d", &account_id, &parent_id);
	     fscanf(stream, "%d %d",
		    &(acc_table[account_id - 1]->reversed), &(acc_table[account_id - 1]->unlimited));
	     data_read_rest_of_line(stream, acc_table[account_id - 1]->name, ACCOUNT_NAME_SIZE);
	     
	     acc_table[account_id - 1]->parent = (parent_id == 0) ? NULL : acc_table[parent_id - 1];
	     acc_table[account_id - 1]->tree_node = NULL;
	}

	acc_list = g_list_reverse(acc_list);
    }
    
    /* and here we read recorded transactions */
    {
	int trans_number, i;
	
	if (fscanf(stream, "%d\n", &trans_number) != EOF) { /* number of them */
	     for (i = 0; i < trans_number; i++) {	/* and the transactions themselves */
		  TransactionData *tmp_trans;
		  int from_id, to_id;
		  int year, month, day;
		  
		  tmp_trans = g_new(TransactionData, 1);
		  trans_list = g_list_prepend(trans_list, tmp_trans);
		  
		  fscanf(stream, "%d %d %lf %d/%d/%d",
			 &from_id, &to_id, 
			 &(tmp_trans->amount),
			 &month, &day, &year);
		  tmp_trans->amount /= 100.0;
		  data_read_rest_of_line(stream, tmp_trans->desc, TRANSACTION_DESC_SIZE);
		  
		  tmp_trans->acc_from = acc_table[from_id - 1];
		  tmp_trans->acc_to = acc_table[to_id - 1];
		  g_date_set_dmy(&(tmp_trans->date), day, month, year);

		  tmp_trans->inbound = TRUE;	/* all transactions are not limited at the begining */
		  tmp_trans->row = (-1);
	     }

	     trans_list = g_list_reverse(trans_list);
	}
    }

    /* and here we read scheduled transactions */
    {
	 int schedule_number, i;

	 if (fscanf(stream, "%d\n", &schedule_number) != EOF) {	/* number of them */
	      for (i = 0; i < schedule_number; i++) {	/* and the transactions themselves */
		   ScheduleData *tmp_schedule;
		   int from_id, to_id;
		   int year, month, day;

		   tmp_schedule = g_new(ScheduleData, 1);
		   schedule_list = g_list_prepend(schedule_list, tmp_schedule);
		   
		   fscanf(stream, "%d %d %d %d %d %lf %d/%d/%d ",
			  &from_id, &to_id, 
			  &(tmp_schedule->frequency),
        &(tmp_schedule->factor),
        &(tmp_schedule->prompt), 
			  &(tmp_schedule->amount),
			  &month, &day, &year);

		   tmp_schedule->amount /= 100.0;
		   data_read_rest_of_line(stream, tmp_schedule->desc, SCHEDULE_DESC_SIZE);
		   
		   tmp_schedule->acc_from = acc_table[from_id - 1];
		   tmp_schedule->acc_to = acc_table[to_id - 1];
		   
		   g_date_set_dmy(&(tmp_schedule->date), day, month, year);

		   tmp_schedule->row = (-1);
	    }

	    schedule_list = g_list_reverse(schedule_list);
	}
    }

    g_free(acc_table);

    if (fclose(stream)) {
	 g_printerr("Can't close stream to %s\n", filename);
    }
}

void data_write(const gchar *filename)
{
    FILE *stream;
    GList *cursor;
    gint counter;

    stream = fopen(filename, "w");
    if (stream == NULL) {
	 g_printerr("Failed to open %s\nFile not saved\n", filename);
	 return;
    }

    /* so now we have to count how many accounts are there */
    counter = 0;
    for (cursor = acc_list; cursor != NULL; cursor = g_list_next(cursor)) {
	 counter++;
	 AccountPtr(cursor->data)->extra_data = GINT_TO_POINTER(counter);
    }

    /* and now we can write down the accounts */
    {
	AccountData *account;
	fprintf(stream, "%d\n", g_list_length(acc_list));
	for (cursor = acc_list; cursor != NULL; cursor = g_list_next(cursor)) {
	    account = AccountPtr(cursor->data);
	    fprintf(stream, "%d %d %d %d %s\n",
		    GPOINTER_TO_INT(account->extra_data),
		    account->parent == NULL ? 0 : GPOINTER_TO_INT(account->parent->extra_data),
		    account->reversed, account->unlimited, account->name);
	}
    }

    /* now we write down the transactionss */
    {
	TransactionData *transaction;

	fprintf(stream, "%d\n", g_list_length(trans_list));
	for (cursor = trans_list; cursor != NULL; cursor = g_list_next(cursor)) {
	    transaction = TransactionPtr(cursor->data);
	    fprintf(stream, "%3d %3d %10.0f %d/%d/%d %s\n",
		    GPOINTER_TO_INT(transaction->acc_from->extra_data),
		    GPOINTER_TO_INT(transaction->acc_to->extra_data),
		    transaction->amount * 100,
		    g_date_month(&(transaction->date)), 
		    g_date_day(&(transaction->date)), 
		    g_date_year(&(transaction->date)),
		    transaction->desc);
	}
    }

    /* write schedule */
    {
	 ScheduleData *schedule;
	 
	 fprintf(stream, "%d\n", g_list_length(schedule_list));
	 for (cursor = schedule_list; cursor != NULL; cursor = g_list_next(cursor)) {
	      schedule = SchedulePtr(cursor->data);
	      fprintf(stream, "%3d %3d %3d %3d %3d %10.0f %d/%d/%d %s\n",
		      GPOINTER_TO_INT(schedule->acc_from->extra_data),
		      GPOINTER_TO_INT(schedule->acc_to->extra_data),
		      schedule->frequency,
		      schedule->factor,
		      schedule->prompt,
		      schedule->amount * 100,
		      g_date_month(&(schedule->date)), 
		      g_date_day(&(schedule->date)), 
		      g_date_year(&(schedule->date)),
		      schedule->desc);
	 }
    }    
    
    if (fclose(stream)) {
	 g_printerr("Can't close stream to %s. Disk full?\n", filename);
    }
}

/******************************
 * REBOUNDING & RECALCULATING *
 ******************************/

void data_rebound_transactions(const gchar *expr)
{
     GList *cursor;

     py_module_compile_limit(expr);
     if (py_module_error_occurred())
	  return;

     for (cursor = trans_list; cursor != NULL; cursor = g_list_next(cursor)) {
	  TransactionData *trans = TransactionPtr(cursor->data);
	  trans->inbound = py_module_eval_limit(trans);
	  if (py_module_error_occurred())
	       break;
     }

     py_module_release_compiled_limit();
}

void data_recalculate_accounts(void)
{
    GList *cursor;
    AccountData *account;
    TransactionData *transaction;

    /* first we need to set all account's sums to 0 */
    for (cursor = acc_list; cursor != NULL; cursor = g_list_next(cursor))
	AccountPtr(cursor->data)->amount = 0;

    /* and now we go tyhrough all transactions modyfying account's amounts as needed */
    for (cursor = trans_list; cursor != NULL; cursor = g_list_next(cursor)) {
	transaction = TransactionPtr(cursor->data);
	for (account = transaction->acc_from; account != NULL; account = account->parent)
	     if (transaction->inbound || account->unlimited) {
		  if (account->reversed)
		       account->amount += transaction->amount;
		  else
		       account->amount -= transaction->amount;
	     }

	for (account = transaction->acc_to; account != NULL; account = account->parent)
	     if (transaction->inbound || account->unlimited) {
		  if (account->reversed)
		       account->amount -= transaction->amount;
		  else
		       account->amount += transaction->amount;
	     }
    }
}

/******************************
 *  INITIALIZATION & DONE     *
 ******************************/

void data_init(void)
{
    acc_list = NULL;
    trans_list = NULL;
    schedule_list = NULL;
}

void data_done(void)
{
    GList *cursor;

    for (cursor = acc_list; cursor != NULL; cursor = g_list_next(cursor))
	g_free(cursor->data);
    g_list_free(acc_list);

    for (cursor = trans_list; cursor != NULL; cursor = g_list_next(cursor))
	g_free(cursor->data);
    g_list_free(trans_list);

    for(cursor = schedule_list; cursor != NULL; cursor = g_list_next(cursor))
    	g_free(cursor->data);
    g_list_free(schedule_list);
}

/******************************
 *   STANDARD ACCOUNTS TREE   *
 ******************************/

AccountData *data_account_new(const gchar *name, 
			      const gboolean reversed, const gboolean unlimited, 
			      AccountData *parent)
{
     AccountData *account;
     account = g_new(AccountData, 1);
     strcpy(account->name, name);
     account->reversed = reversed;
     account->unlimited = unlimited;
     account->parent = parent;     

     account->amount = 0;
     account->tree_node = NULL;

     return account;
}

void data_set_standard_accounts(void)
{
     AccountData *account[100];
     int i;

     for (i = 0; i < 100; i++)
	  account[i] = NULL;

     account[ 0] = data_account_new(_("Expenses"), TRUE, FALSE, NULL); 
     account[ 1] = data_account_new(_("Car"), TRUE, FALSE, account[0]); 
     account[ 2] = data_account_new(_("Car service"), TRUE, FALSE, account[1]); 
     account[ 3] = data_account_new(_("Fuel"), TRUE, FALSE, account[1]); 
     account[ 4] = data_account_new(_("Misc (car)"), TRUE, FALSE, account[1]); 
     account[ 5] = data_account_new(_("Bank charge"), TRUE, FALSE, account[0]); 
     account[ 6] = data_account_new(_("Interest paid"), TRUE, FALSE, account[5]); 
     account[ 7] = data_account_new(_("Bank charges"), TRUE, FALSE, account[5]); 
     account[ 8] = data_account_new(_("Bills"), TRUE, FALSE, account[0]); 
     account[ 9] = data_account_new(_("Phone"), TRUE, FALSE, account[8]); 
     account[10] = data_account_new(_("Radio & TV"), TRUE, FALSE, account[8]); 
     account[11] = data_account_new(_("Electricity"), TRUE, FALSE, account[8]); 
     account[12] = data_account_new(_("Garbage"), TRUE, FALSE, account[8]); 
     account[13] = data_account_new(_("House rent"), TRUE, FALSE, account[8]); 
     account[14] = data_account_new(_("Water/sewer"), TRUE, FALSE, account[8]); 
     account[15] = data_account_new(_("Gas"), TRUE, FALSE, account[8]); 
     account[16] = data_account_new(_("Child care"), TRUE, FALSE, account[0]); 
     account[17] = data_account_new(_("Education"), TRUE, FALSE, account[0]); 
     account[18] = data_account_new(_("Books (education)"), TRUE, FALSE, account[17]); 
     account[19] = data_account_new(_("Fees (education)"), TRUE, FALSE, account[17]); 
     account[20] = data_account_new(_("Tuition"), TRUE, FALSE, account[17]); 
     account[21] = data_account_new(_("Health care"), TRUE, FALSE, account[0]); 
     account[22] = data_account_new(_("Dental care"), TRUE, FALSE, account[21]); 
     account[23] = data_account_new(_("Vision care"), TRUE, FALSE, account[21]); 
     account[24] = data_account_new(_("Hospital"), TRUE, FALSE, account[21]); 
     account[25] = data_account_new(_("Physician"), TRUE, FALSE, account[21]); 
     account[26] = data_account_new(_("Medicaments"), TRUE, FALSE, account[21]); 
     account[27] = data_account_new(_("Home"), TRUE, FALSE, account[0]); 
     account[28] = data_account_new(_("Maintenance (home)"), TRUE, FALSE, account[27]); 
     account[29] = data_account_new(_("Furnishing"), TRUE, FALSE, account[27]); 
     account[30] = data_account_new(_("Misc (home)"), TRUE, FALSE, account[27]); 
     account[31] = data_account_new(_("Personal"), TRUE, FALSE, account[0]); 
     account[32] = data_account_new(_("Entertainment"), TRUE, FALSE, account[31]); 
     account[33] = data_account_new(_("Books/magazines"), TRUE, FALSE, account[32]); 
     account[34] = data_account_new(_("Cinema"), TRUE, FALSE, account[32]); 
     account[35] = data_account_new(_("Music"), TRUE, FALSE, account[32]); 
     account[36] = data_account_new(_("Theatre"), TRUE, FALSE, account[32]); 
     account[37] = data_account_new(_("Computer games"), TRUE, FALSE, account[32]); 
     account[38] = data_account_new(_("Misc (entertainment)"), TRUE, FALSE, account[32]); 
     account[39] = data_account_new(_("Food"), TRUE, FALSE, account[31]); 
     account[40] = data_account_new(_("Gifts"), TRUE, FALSE, account[31]); 
     account[41] = data_account_new(_("Computer"), TRUE, FALSE, account[31]); 
     account[42] = data_account_new(_("Pet care"), TRUE, FALSE, account[0]); 
     account[43] = data_account_new(_("Food (pet)"), TRUE, FALSE, account[42]); 
     account[44] = data_account_new(_("Supplies (pet)"), TRUE, FALSE, account[42]); 
     account[45] = data_account_new(_("Veterinarian"), TRUE, FALSE, account[42]); 
     account[46] = data_account_new(_("Taxes"), TRUE, FALSE, account[0]); 
     account[47] = data_account_new(_("Misc (expenses)"), TRUE, FALSE, account[0]); 

     account[50] = data_account_new(_("Income"), TRUE, FALSE, NULL); 
     account[51] = data_account_new(_("Salary"), TRUE, FALSE, account[50]); 
     account[52] = data_account_new(_("Child support"), TRUE, FALSE, account[50]); 
     account[53] = data_account_new(_("Gifts received"), TRUE, FALSE, account[50]); 
     account[54] = data_account_new(_("Lotteries"), TRUE, FALSE, account[50]); 
     account[55] = data_account_new(_("Tax refund"), TRUE, FALSE, account[50]); 
     account[56] = data_account_new(_("Misc (income)"), TRUE, FALSE, account[50]); 

     account[60] = data_account_new(_("Assets"), FALSE, TRUE, NULL); 
     account[61] = data_account_new(_("Checking account"), FALSE, TRUE, account[60]); 
     account[62] = data_account_new(_("Cash"), FALSE, TRUE, account[60]); 
     account[63] = data_account_new(_("Reconciled"), FALSE, TRUE, account[61]); 

     account[70] = data_account_new(_("Opening balance"), TRUE, TRUE, NULL);

     for (i = 0; i < 100; i++)
	  if (account[i] != NULL)
	       acc_list = g_list_prepend(acc_list, account[i]);
}
