#include <config.h>

#include <errno.h>
#include <signal.h>
#include <unistd.h> /* used in subprocessing */
#include <sys/wait.h>

#include "actions.h"
#include "data.h"

#include "report_acc_axis.h"
#include "report_grp_axis.h"
#include "report_progress.h"
#include "report.h"

static void report_sigchld_handler(int signum);
static sig_t report_prev_sigchld_handler;

/*
 * INIT & DONE *************************************************************
 */

void report_init(void)
{
  report_acc_axis_init();
  report_grp_axis_init();

  /* install singal handler to get rid of zombie-children */
  report_prev_sigchld_handler = signal(SIGCHLD, report_sigchld_handler);
  if (report_prev_sigchld_handler == SIG_ERR)
    g_printerr(_("Couldn't install a handler for sigchld signal\n"));
}

void report_done(void)
{
  report_acc_axis_done();
  report_grp_axis_done();

  /* restore a previous signal handler */
  signal(SIGCHLD, report_prev_sigchld_handler);
}

/*
 * (RE)CREATION OF THE REPORT CLIST WIDGET *********************************
 */

static GtkWidget *report_widget = NULL;
static GtkWidget *report_widget_container = NULL;

GtkWidget *report_widget_new(gint groups_no)
{
  gint i;
     
  if (report_widget_container == NULL) {
    report_widget_container = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy
      (GTK_SCROLLED_WINDOW(report_widget_container), 
       GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
    gtk_widget_show(report_widget_container);
  } else {
    g_assert(report_widget != NULL);
    gtk_container_remove(GTK_CONTAINER(report_widget_container),report_widget);
  }

  report_widget = gtk_clist_new(groups_no + 1);
  gtk_clist_column_titles_hide(GTK_CLIST(report_widget));
  for (i = 0; i < groups_no + 1; i++)
    gtk_clist_set_column_auto_resize(GTK_CLIST(report_widget), i, TRUE);
  gtk_clist_set_shadow_type(GTK_CLIST(report_widget), GTK_SHADOW_ETCHED_IN);
  gtk_widget_show(report_widget);
  
  gtk_container_add(GTK_CONTAINER(report_widget_container), report_widget);
     
  return report_widget_container;
}

/*
 * CREATION & DESTRUCTION OF TEMPORARY REPORT-TABLES ********************
 */

static gdouble **report_table_new(gint accounts_no, gint groups_no)
{
  gdouble **report_table;
  gint i;

  accounts_no = g_list_length(report_acc_axis_list);
  groups_no = g_list_length(report_grp_axis_list);

  report_table = g_new(gdouble *, groups_no);
  for (i = 0; i < groups_no; i++)
    report_table[i] = g_new0(gdouble, accounts_no);
  
  return report_table;
}

static void report_table_free(gdouble **report_table, gint groups_no)
{
  gint i;
  for (i = 0; i < groups_no; i++)
    g_free(report_table[i]);
  g_free(report_table);
}

/*
 * FUNCTIONS FOR FILLING IN TEMPORARY REPORT TABLES **********************
 * (they invoke progress dialog)
 */

/* returns TRUE on success */
static gboolean report_table_calculate_grp(gdouble *report_table_grp, 
					   gchar *expr)
{
  GList *acc_cursor, *trans_cursor, *rad_cursor;
  gint i;

  /* set all accounts to 0.0 */
  for (acc_cursor = acc_list; 
       acc_cursor != NULL; 
       acc_cursor = g_list_next(acc_cursor))
    AccountPtr(acc_cursor->data)->amount = 0.0;

  /* compile the group-limit expression */
  py_module_compile_limit(expr);
  if (py_module_error_occurred())
    return FALSE;

  /* calculate accounts' amounts for limit=main_limit + group_expr */
  for (trans_cursor = trans_list; 
       trans_cursor != NULL; 
       trans_cursor = g_list_next(trans_cursor)) {
      
    TransactionData *transaction;
    AccountData *account;
    gboolean report_inbound;

    transaction = TransactionPtr(trans_cursor->data);
    report_inbound = 
      transaction->inbound && py_module_eval_limit(transaction);
    if (py_module_error_occurred()) {
      py_module_release_compiled_limit();
      return FALSE;
    }

    for (account = transaction->acc_from; 
	 account != NULL; 
	 account = account->parent)
      if (report_inbound || account->unlimited) 
	account->amount -= transaction->amount;

    for (account = transaction->acc_to; 
	 account != NULL; 
	 account = account->parent)
      if (report_inbound || account->unlimited) 
	account->amount += transaction->amount;
  }
  py_module_release_compiled_limit();
    
  /* copy calculated accounts' amounts to report_table */
  for (rad_cursor = report_acc_axis_list, i = 0; 
       rad_cursor != NULL; 
       rad_cursor = g_list_next(rad_cursor), i++) {
    ReportAccountData *rad;
    AccountData *acc;

    rad = ReportAccountDataPtr(rad_cursor->data);
    acc = rad->account;

    report_table_grp[i] = (rad->reversed ? -1 : 1) * acc->amount;
  }

  return TRUE;
}

/* returns TRUE on success */
static gboolean report_table_calculate(gdouble **report_table)
{
  GList *gr_cursor;
  gint i;

  report_progress_init(g_list_length(report_grp_axis_list));
  for (gr_cursor = report_grp_axis_list, i = 0; 
       gr_cursor != NULL;
       gr_cursor = g_list_next(gr_cursor), i++) {

    if (!report_table_calculate_grp(report_table[i], 
			    ReportGroupDataPtr(gr_cursor->data)->expr)) {
      report_progress_done();
      return FALSE;
    }

    report_progress_set_group(i+1);
    while (gtk_events_pending())
      gtk_main_iteration();

    if (report_progress_cancel_now()) {
      report_progress_done();
      return FALSE;
    }
  }

  /* restore proper values of accounts */
  data_recalculate_accounts();

  /* finish and tell that everything went OK */
  report_progress_done();
  return TRUE;
}

/*
 * RECALCULATING OF REPORT CLIST ****************************************
 */

void report_rebuild(void)
{
  GList *cursor;   

  gdouble **report_table;
  gint accounts_no, groups_no;
  gint i, j;

  gchar **text;

  report_acc_axis_check();

  accounts_no = g_list_length(report_acc_axis_list);
  groups_no = g_list_length(report_grp_axis_list);
  report_table = report_table_new(accounts_no, groups_no);
  if (!report_table_calculate(report_table)) {
    report_table_free(report_table, groups_no);
    return; 
  }

  report_widget_new(groups_no);

  gtk_clist_freeze(GTK_CLIST(report_widget));
  gtk_clist_clear(GTK_CLIST(report_widget));

  /* Groups titles */
  text = g_new0(gchar *, groups_no + 2);
  text[0] = g_strdup("");
  for (cursor = report_grp_axis_list, i = 0; 
       cursor != NULL; 
       cursor = g_list_next(cursor), i++) {

    text[i+1] = g_strdup(ReportGroupDataPtr(cursor->data)->label);
  }
  gtk_clist_append(GTK_CLIST(report_widget), text);
  g_strfreev(text);

  /* Accounts names, and the calculated data */
  for (cursor = report_acc_axis_list, i = 0; 
       cursor != NULL; 
       cursor = g_list_next(cursor), i++) {

    text = g_new0(gchar *, groups_no + 2);
    text[0] = g_strdup(ReportAccountDataPtr(cursor->data)->account->name);
    for (j = 0; j < groups_no; j++)
      text[j + 1] = g_strdup_printf("%.2lf", report_table[j][i]);
    
    gtk_clist_append(GTK_CLIST(report_widget), text);
    g_strfreev(text);
  }

  gtk_clist_thaw(GTK_CLIST(report_widget));

  report_table_free(report_table, groups_no);
}

/*
 * REPORT EXPORT FUNCTIONS *************************************************
 */

static void report_write_to_stream(FILE *stream)
{
  gint accounts_no, groups_no;
  gint acc, grp;
  gchar *text;

  fprintf(stream, "# Emma report output stream\n");

  accounts_no = GTK_CLIST(report_widget)->rows - 1 ;
  groups_no = GTK_CLIST(report_widget)->columns - 1;

  fprintf(stream, "%s", "FIXME (title)\n");
  
  fprintf(stream, "%d %d\n", accounts_no, groups_no);
     
  for (acc = 0; acc < accounts_no; acc++) {
    gtk_clist_get_text(GTK_CLIST(report_widget), acc + 1, 0, &text);
    fprintf(stream, "%s\n", text);
  }

  for (grp = 0; grp < groups_no; grp++) {
    gtk_clist_get_text(GTK_CLIST(report_widget), 0, grp + 1, &text);
    fprintf(stream, "%s\n", text);
  }

  for (acc = 0; acc < accounts_no; acc++) {
    for (grp = 0; grp < groups_no; grp++) {
      gtk_clist_get_text(GTK_CLIST(report_widget), acc + 1, grp + 1, &text);
      fprintf(stream, "%s", text);
      if (grp != (groups_no-1))
	fprintf(stream, "\t");
    }
    fprintf(stream, "\n");
  }
}

static void report_sigchld_handler(int signum)
{
  int pid, status, serrno;
  serrno = errno;
  while (1) {
    pid = waitpid (WAIT_ANY, &status, WNOHANG);
    if (pid < 0) {
      if (errno != ECHILD)
	perror ("waitpid");
      break;
    }
    if (pid == 0)
      break;

    /* notice_termination (pid, status); */
  }
  errno = serrno;
}

void report_filter(const gchar *command)
{
  pid_t pid;
  int input_pipe[2];
     
  if (pipe(input_pipe)) {
    action_error_msg(_("Failed to create a pipe for the subprocess"),  NULL);
    return;
  }
     
  pid = fork();
  if (pid == (pid_t) 0) {
    /* This is the child process. */
    dup2(input_pipe[0], STDIN_FILENO);
    close(input_pipe[0]);
    close(input_pipe[1]);
    
    execl("/bin/sh", "/bin/sh", "-c", command, NULL);

    g_printerr("emma: Failed to execute the shell!!!???");
    exit(-1);
  } else if (pid > (pid_t) 0) {
    /* This is the parent process. */
    FILE *childs_input;

    close(input_pipe[0]);
    
    childs_input = fdopen(input_pipe[1], "w");
    report_write_to_stream(childs_input);
    fclose(childs_input);
  } else {
    /* The fork failed. */
    close(input_pipe[0]);
    close(input_pipe[1]);
    action_error_msg(_("Failed to fork() the subprocess"), NULL);
  }
}


