#include <config.h>

#include <langinfo.h>
#include <math.h>

#include "date_entry.h"
#include "gnome-ucombo.h"
#include "misc.h"

//#include "report_acc_axis.h"
#include "report_grp_axis.h"

#include "report_druid.h"

/*
0123456789012345678901234567890123456789012345678901234567890123456789012345678
*/

/*
 * - every DruidPage has its structure named by 
 *   adding page_ prefix before its name
 *   (the struct. should contain GnomeDruidPage *page field)
 * - every DruidPage should connect its 'cancel'
 *   event with report_druid_hide callback function
 * - to control the flow in the druid you can use
 *   report_druid_switch_page callback with
 *   user data set to the page you want to switch to
 * 
 * The schema below shows how the flow in the druid is organised
 *
 * start 
 *   |   
 * choice---(weekday,month)---\
 *   |                        |
 *   |(time period)           |
 *   |                        |
 * period_df                /
 *   |                      /
 * period_lab<-not impl.   /
 *   |                    /
 *   +-------------------/
 *   |
 * finish
 *
 */


/* 
 * GLOBAL VARS FOR DRUID REFERENCE  ****************************************
 */

static struct _Druid {
  GtkWidget *druid;
  GtkWidget *window;
} druid;

static gboolean druid_already_built = FALSE;



/* 
 * DRUID CONTROL-FLOW FUNCTIONS **********************************************
 */

static gboolean report_druid_finish(GtkWidget* , GtkWidget* , gpointer);
static GnomeDruidPage *report_druid_get_first_page(void);

static gboolean report_druid_hide(GtkWidget *widget, 
				  GtkWidget *event, 
				  gpointer callback_data)
{
  g_return_val_if_fail(druid_already_built, FALSE);

  gtk_widget_hide(druid.window);
  gnome_druid_set_page(GNOME_DRUID(druid.druid),report_druid_get_first_page());

  return TRUE;
}

static gboolean report_druid_switch_page(GtkWidget *widget, 
					 GtkWidget *event, 
					 gpointer page)
{
  g_return_val_if_fail(druid_already_built, FALSE);
  g_assert(GNOME_IS_DRUID_PAGE(page));

  gnome_druid_set_page(GNOME_DRUID(druid.druid), GNOME_DRUID_PAGE(page));
  return TRUE;
}

static void report_druid_set_back(GnomeDruidPage *from, GnomeDruidPage *to)
{
  gtk_signal_connect_after(GTK_OBJECT(from), "back", 
			   GTK_SIGNAL_FUNC(report_druid_switch_page), to);
}

static void report_druid_set_next(GnomeDruidPage *from, GnomeDruidPage *to)
{
  gtk_signal_connect_after(GTK_OBJECT(from), "next", 
			   GTK_SIGNAL_FUNC(report_druid_switch_page), to);
}



/* 
 * PAGE: END POINTS *******************************************************
 */

static struct _PageStart {
  GtkWidget *page;
} page_start;

static struct _PageFinish {
  GtkWidget *page;
} page_finish;

static GnomeDruidPage *report_druid_get_first_page(void)
{
  return GNOME_DRUID_PAGE(page_start.page);
}

static void report_druid_build_start(void)
{
  page_start.page = gnome_druid_page_start_new();
  gnome_druid_page_start_set_title(GNOME_DRUID_PAGE_START(page_start.page), 
				   _("Introduction"));
  gnome_druid_page_start_set_text(GNOME_DRUID_PAGE_START(page_start.page), 
       _("Hello, I am a report creation druid.\n"
	 "\n"
	 "With my help you will have your report ready in notime.\n"
	 "Now click \"Next\" button to proceed."
	 )
				  );
  gnome_druid_page_start_set_logo_bg_color(
	     GNOME_DRUID_PAGE_START(page_start.page),
	   &(GNOME_DRUID_PAGE_START(page_start.page)->background_color));

  gtk_widget_show(page_start.page);
  gnome_druid_append_page(GNOME_DRUID(druid.druid), 
			  GNOME_DRUID_PAGE(page_start.page));
  gtk_signal_connect(GTK_OBJECT(page_start.page), "cancel", 
		     GTK_SIGNAL_FUNC(report_druid_hide), NULL);
} 

static void report_druid_build_finish(void)
{
  page_finish.page = gnome_druid_page_finish_new();
  gnome_druid_page_finish_set_title(GNOME_DRUID_PAGE_FINISH(page_finish.page), 
				    _("Let's go"));
  gnome_druid_page_finish_set_text(GNOME_DRUID_PAGE_FINISH(page_finish.page), 
       _("Ok - everything seems to be ready for report creation.\n"
	 "\n"
	 "If you want the report to be created - press \"Finish\" button\n"
	 "If anything needs correction - press \"Back\" button.\n"
	 )
				  );
  gnome_druid_page_finish_set_logo_bg_color(
	     GNOME_DRUID_PAGE_FINISH(page_finish.page),
	   &(GNOME_DRUID_PAGE_FINISH(page_finish.page)->background_color));

  gtk_widget_show(page_finish.page);
  gnome_druid_append_page(GNOME_DRUID(druid.druid), 
			  GNOME_DRUID_PAGE(page_finish.page));
  gtk_signal_connect(GTK_OBJECT(page_finish.page), "finish", 
		   GTK_SIGNAL_FUNC(report_druid_finish), NULL);
  gtk_signal_connect(GTK_OBJECT(page_finish.page), "cancel", 
		     GTK_SIGNAL_FUNC(report_druid_hide), NULL);
}

/* 
 * PAGE: PERIOD DEF *******************************************************
 */

static struct _PagePeriodDf {
  GtkWidget *page;

  GtkAdjustment *groups_no_adj;
  GtkWidget *groups_no;

  GtkAdjustment *period_len_adj;
  GtkWidget *start_date, *period_len, *end_date;
  GtkWidget *sd_radio, *pl_radio, *ed_radio;
  GtkWidget *day_radio, *mon_radio;
  gint calc_what;

  GtkAdjustment *incl_label_adj;
  GtkWidget *incl_label;
  GtkWidget *label_entry;
} page_period_df;

static void gtk_spin_button_my_update(GtkSpinButton *spin)
{
  GtkAdjustment *adj;
  gfloat value;

  adj = gtk_spin_button_get_adjustment(spin);
  sscanf(gtk_entry_get_text(GTK_ENTRY(spin)), "%f", &value);
  
  if ((adj->lower <= value) && (value <= adj->upper))
    adj->value = value; //gtk_spin_button_set_value(spin, value);
}

/* callback for pressing one of "calculate from others" radio buttons */
static void page_period_df_calc_what(GtkWidget *widget, gpointer callback_data)
{
  gint what;

  page_period_df.calc_what = GPOINTER_TO_INT(callback_data);
  what = page_period_df.calc_what;

  gtk_widget_set_sensitive(page_period_df.start_date, what != 1);
  gtk_widget_set_sensitive(page_period_df.end_date, what != 2);
  gtk_widget_set_sensitive(page_period_df.period_len, what != 3);
}

static void page_period_df_calculate(GtkWidget *widget, gpointer callback_data)
{
  GDate start, end;
  gint groups_no;
  gfloat length;
  gint what = page_period_df.calc_what;

  /* this condition is needed to check against infinite loop */
  if (what == GPOINTER_TO_INT(callback_data))
    return;

  /* if spin button entries changed - update */
  switch (GPOINTER_TO_INT(callback_data)) {
  case 0:
    gtk_spin_button_my_update(GTK_SPIN_BUTTON(page_period_df.groups_no));
    break;
  case 3:
    gtk_spin_button_my_update(GTK_SPIN_BUTTON(page_period_df.period_len));
    break;
  }

  /* read current values from entries
   * (if anything is invalid return without touching anything) */
  groups_no = page_period_df.groups_no_adj->value;
  length = page_period_df.period_len_adj->value * groups_no;
  if(((what!=1) && (date_entry_get_date(page_period_df.start_date, &start)))
  || ((what!=2) && (date_entry_get_date(page_period_df.end_date, &end))) )
    return;

  /* calculate selected entry from the other two and groups_no */
  switch (what) {
  case 1:
    start = end;
    if (GTK_TOGGLE_BUTTON(page_period_df.day_radio)->active) {
      g_date_subtract_days(&start, length);
    } else {
      g_date_subtract_months(&start, floor(length));
      g_date_subtract_days(&start, floor(30 * (length-floor(length))));
    }
    date_entry_set_date(page_period_df.start_date, &start);
    break;

  case 2:
    end = start;
    if (GTK_TOGGLE_BUTTON(page_period_df.day_radio)->active) {
      g_date_add_days(&end, length);
    } else {
      g_date_add_days(&end, floor(30 * (length-floor(length))));
      g_date_add_months(&end, floor(length));
    }
    date_entry_set_date(page_period_df.end_date, &end);
    break;

  case 3:
    if (GTK_TOGGLE_BUTTON(page_period_df.day_radio)->active) {
      length = g_date_julian(&end) - g_date_julian(&start);
      length /= groups_no;
    } else {
      length = 0;
      length += ((g_date_year(&end) - g_date_year(&start)) * 12);
      length += (g_date_month(&end) - g_date_month(&start));
      length += ((g_date_day(&end) - g_date_day(&start)) / 30.0);
      length /= groups_no;
    }
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(page_period_df.period_len), 
			      length);
    break;
  }
}

static gboolean page_period_df_conditional_next(GtkWidget *widget,
						GtkWidget *event, 
						gpointer callback_data)
{
  GDate date;
  
  g_date_clear(&date, 1);

  if((date_entry_get_date(page_period_df.start_date, &date))
  || (date_entry_get_date(page_period_df.end_date, &date))      ) {
    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "next");
    return TRUE;
  }

  gtk_spin_button_my_update(GTK_SPIN_BUTTON(page_period_df.groups_no));
  gtk_spin_button_my_update(GTK_SPIN_BUTTON(page_period_df.period_len));
  gtk_spin_button_update(GTK_SPIN_BUTTON(page_period_df.incl_label));

  /* everything ok */
  return FALSE;
}

static void report_druid_build_period_df(void)
{
  GtkWidget *table, *vbox, *label;
  GDate today;

  page_period_df.page = gnome_druid_page_standard_new();
  gnome_druid_page_standard_set_title(
           GNOME_DRUID_PAGE_STANDARD(page_period_df.page), _("Time periods"));
  gnome_druid_page_standard_set_logo_bg_color(
	    GNOME_DRUID_PAGE_STANDARD(page_period_df.page),
	 &(GNOME_DRUID_PAGE_STANDARD(page_period_df.page)->background_color));

  /* fill in the page */
  table = gtk_table_new(6, 4, FALSE);

  vbox = GNOME_DRUID_PAGE_STANDARD(page_period_df.page)->vbox;
  gtk_box_pack_start(GTK_BOX(vbox), table, TRUE, TRUE, 0);

  /* number of time periods */

  label = gtk_label_new(_("Number of time periods:"));
  gtk_table_attach(GTK_TABLE(table), label, 0, 2, 0, 1, 0, 0, 3, 3);

  page_period_df.groups_no_adj = 
    GTK_ADJUSTMENT(gtk_adjustment_new(10, 1, 10000, 1, 10, 10));
  page_period_df.groups_no=gtk_spin_button_new(page_period_df.groups_no_adj,
						1,0);
  gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(page_period_df.groups_no), TRUE);
  gtk_spin_button_set_update_policy(GTK_SPIN_BUTTON(page_period_df.groups_no),
				    GTK_UPDATE_IF_VALID);
  gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(page_period_df.groups_no),
				    FALSE);
  gtk_signal_connect(GTK_OBJECT(page_period_df.groups_no), "changed",
		     GTK_SIGNAL_FUNC(page_period_df_calculate),
		     GINT_TO_POINTER(0)); 
  gtk_table_attach(GTK_TABLE(table), page_period_df.groups_no, 
		   2, 4, 0, 1, 0, 0, 3, 3);

  /* start date */
  label = gtk_label_new(_("Start date:"));
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, 0, 0, 3, 15);

  page_period_df.start_date = date_entry_new();
  gtk_signal_connect(
                  GTK_OBJECT(GNOME_UCOMBO(page_period_df.start_date)->entry), 
		  "changed", GTK_SIGNAL_FUNC(page_period_df_calculate), 
		  GINT_TO_POINTER(1)); 
  gtk_table_attach(GTK_TABLE(table), page_period_df.start_date, 
		   1, 5, 1, 2, GTK_FILL, 0, 3, 3);

  /* end date */
  label = gtk_label_new(_("End date:"));
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3, 0, 0, 3, 3);

  g_date_set_current(&today);
  page_period_df.end_date = date_entry_new();
  date_entry_set_date(page_period_df.end_date, &today);
  gtk_signal_connect(
		  GTK_OBJECT(GNOME_UCOMBO(page_period_df.end_date)->entry), 
		  "changed", GTK_SIGNAL_FUNC(page_period_df_calculate),
		  GINT_TO_POINTER(2)); 
  gtk_table_attach(GTK_TABLE(table), page_period_df.end_date, 
		   1, 5, 2, 3, GTK_FILL, 0, 3, 3);

  /* period length */
  label = gtk_label_new(_("Period length:"));
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 3, 4, 0, 0, 3, 3);

  page_period_df.period_len_adj = 
    GTK_ADJUSTMENT(gtk_adjustment_new(10000, 0, 10000, 0.1, 1, 1));
  page_period_df.period_len=gtk_spin_button_new(page_period_df.period_len_adj,
						 0.1, 2);
  gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(page_period_df.period_len),TRUE);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(page_period_df.period_len), 5);
  gtk_spin_button_set_update_policy(GTK_SPIN_BUTTON(page_period_df.period_len),
				    GTK_UPDATE_IF_VALID);
  gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(page_period_df.period_len),
				    FALSE);
  gtk_signal_connect(GTK_OBJECT(page_period_df.period_len), "changed",
		     GTK_SIGNAL_FUNC(page_period_df_calculate), 
		     GINT_TO_POINTER(3)); 
  gtk_table_attach(GTK_TABLE(table), page_period_df.period_len, 
		   1, 2, 3, 4, 0, 0, 3, 3);

  page_period_df.day_radio = gtk_radio_button_new_with_label(
            NULL, 
	    _("days"));
  gtk_table_attach(GTK_TABLE(table), page_period_df.day_radio, 
		   2, 3, 3, 4, 0, 0, 3, 3);
  gtk_signal_connect(GTK_OBJECT(page_period_df.day_radio), "clicked", 
		     GTK_SIGNAL_FUNC(page_period_df_calculate), 
		     GINT_TO_POINTER(4));
  gtk_toggle_button_set_active(
                     GTK_TOGGLE_BUTTON(page_period_df.day_radio),TRUE);

  label = gtk_label_new("/");
  gtk_table_attach(GTK_TABLE(table), label, 3, 4, 3, 4, 0, 0, 3, 3);

  page_period_df.mon_radio = gtk_radio_button_new_with_label(
            gtk_radio_button_group(GTK_RADIO_BUTTON(page_period_df.day_radio)),
	    _("months"));
  gtk_signal_connect(GTK_OBJECT(page_period_df.mon_radio), "clicked", 
		     GTK_SIGNAL_FUNC(page_period_df_calculate), 
		     GINT_TO_POINTER(5));
  gtk_table_attach(GTK_TABLE(table), page_period_df.mon_radio, 
		   4, 5, 3, 4, 0, 0, 3, 3);

  /* the right radio items bar */
  label = gtk_label_new(_("Calculate\nfrom others:"));
  gtk_table_attach(GTK_TABLE(table), label, 
		   5, 6, 0, 1, 0, 0, 3, 3);

  page_period_df.sd_radio = gtk_radio_button_new_with_label(
            NULL, 
	    "");
  gtk_table_attach(GTK_TABLE(table), page_period_df.sd_radio, 
		   5, 6, 1, 2, 0, 0, 3, 3);
  gtk_signal_connect(GTK_OBJECT(page_period_df.sd_radio), "clicked", 
		     GTK_SIGNAL_FUNC(page_period_df_calc_what), 
		     GINT_TO_POINTER(1));

  page_period_df.ed_radio = gtk_radio_button_new_with_label(
            gtk_radio_button_group(GTK_RADIO_BUTTON(page_period_df.sd_radio)),
	    "");
  gtk_table_attach(GTK_TABLE(table), page_period_df.ed_radio, 
		   5, 6, 2, 3, 0, 0, 3, 3);
  gtk_signal_connect(GTK_OBJECT(page_period_df.ed_radio), "clicked", 
		     GTK_SIGNAL_FUNC(page_period_df_calc_what), 
		     GINT_TO_POINTER(2));

  page_period_df.pl_radio = gtk_radio_button_new_with_label(
            gtk_radio_button_group(GTK_RADIO_BUTTON(page_period_df.sd_radio)),
	    "");
  gtk_table_attach(GTK_TABLE(table), page_period_df.pl_radio, 
		   5, 6, 3, 4, 0, 0, 3, 3);
  gtk_signal_connect(GTK_OBJECT(page_period_df.pl_radio), "clicked", 
		     GTK_SIGNAL_FUNC(page_period_df_calc_what), 
		     GINT_TO_POINTER(3));

  gtk_button_clicked(GTK_BUTTON(page_period_df.sd_radio));
  page_period_df_calculate(NULL, GINT_TO_POINTER(-1));

  /* horizontal separator */
  label = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);

  /* things controlling labels */
  table = gtk_table_new(4, 2, FALSE);

  vbox = GNOME_DRUID_PAGE_STANDARD(page_period_df.page)->vbox;
  gtk_box_pack_start(GTK_BOX(vbox), table, TRUE, TRUE, 0);

  /* every i'th label */
  label = gtk_label_new(_("Include every"));
  gtk_table_attach(GTK_TABLE(table), label, 0, 2, 0, 1, 0, 0, 3, 3);

  label = gtk_label_new(_("label"));
  gtk_table_attach(GTK_TABLE(table), label, 3, 4, 0, 1, 0, 0, 3, 3);

  page_period_df.incl_label_adj = 
    GTK_ADJUSTMENT(gtk_adjustment_new(10000, 1, 10000, 1, 5, 1));
  page_period_df.incl_label=gtk_spin_button_new(page_period_df.incl_label_adj,
						1, 0);
  gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(page_period_df.incl_label),TRUE);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(page_period_df.incl_label), 1);
  gtk_spin_button_set_update_policy(GTK_SPIN_BUTTON(page_period_df.incl_label),
				    GTK_UPDATE_IF_VALID);
  gtk_spin_button_set_snap_to_ticks(GTK_SPIN_BUTTON(page_period_df.incl_label),
				    TRUE);
  gtk_table_attach(GTK_TABLE(table), page_period_df.incl_label, 
		   2, 3, 0, 1, 0, 0, 3, 3);

  /* label format */
  label = gtk_label_new(_("Label format:"));
  gtk_table_attach(GTK_TABLE(table), label, 0, 2, 1, 2, 0, 0, 3, 3);

  page_period_df.label_entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(page_period_df.label_entry), misc_date_fmt);
  gtk_table_attach(GTK_TABLE(table), page_period_df.label_entry, 
		   2, 4, 1, 2, 0, 0, 3, 3);

  /* show the page and connect to control-flow managing functions */
  gtk_widget_show_all(page_period_df.page);
  gnome_druid_append_page(GNOME_DRUID(druid.druid), 
			  GNOME_DRUID_PAGE(page_period_df.page));
  gtk_signal_connect(GTK_OBJECT(page_period_df.page), "cancel", 
		     GTK_SIGNAL_FUNC(report_druid_hide), NULL);
  gtk_signal_connect(GTK_OBJECT(page_period_df.page), "next", 
		     GTK_SIGNAL_FUNC(page_period_df_conditional_next), NULL);
}

/* 
 * PAGE: CHOICE *******************************************************
 */

static struct _PageChoice {
  GtkWidget *page;

  GtkWidget *period_radio, *weekday_radio, *month_radio;
} page_choice;

static gboolean report_druid_choice_fork(GtkWidget *widget, 
					 GtkWidget *event, 
					 gpointer callback_data)
{
  g_return_val_if_fail(druid_already_built, FALSE);

  if (GTK_TOGGLE_BUTTON(page_choice.period_radio)->active) {
    gnome_druid_set_page(GNOME_DRUID(druid.druid), 
			 GNOME_DRUID_PAGE(page_period_df.page));
  } else if(GTK_TOGGLE_BUTTON(page_choice.weekday_radio)->active) {
    gnome_druid_set_page(GNOME_DRUID(druid.druid), 
			 GNOME_DRUID_PAGE(page_finish.page));
  } else if(GTK_TOGGLE_BUTTON(page_choice.month_radio)->active) {
    gnome_druid_set_page(GNOME_DRUID(druid.druid), 
			 GNOME_DRUID_PAGE(page_finish.page));
  } else
    g_assert_not_reached();

  return TRUE;
}

static void report_druid_build_choice(void)
{
  GtkWidget *vbox;

  page_choice.page = gnome_druid_page_standard_new();
  gnome_druid_page_standard_set_title(
              GNOME_DRUID_PAGE_STANDARD(page_choice.page), _("Report type"));
  gnome_druid_page_standard_set_logo_bg_color(
	     GNOME_DRUID_PAGE_STANDARD(page_choice.page),
	   &(GNOME_DRUID_PAGE_STANDARD(page_choice.page)->background_color));

  /* fill in the page */
  vbox = GNOME_DRUID_PAGE_STANDARD(page_choice.page)->vbox;

  page_choice.period_radio = gtk_radio_button_new_with_label(
            NULL, 
	    _("time periods"));
  gtk_box_pack_start_defaults(GTK_BOX(vbox), page_choice.period_radio); 
  gtk_toggle_button_set_active(
                     GTK_TOGGLE_BUTTON(page_choice.period_radio),TRUE);

  page_choice.weekday_radio = gtk_radio_button_new_with_label(
            gtk_radio_button_group(GTK_RADIO_BUTTON(page_choice.period_radio)),
	    _("by weekday"));
  gtk_box_pack_start_defaults(GTK_BOX(vbox), page_choice.weekday_radio); 

  page_choice.month_radio = gtk_radio_button_new_with_label(
            gtk_radio_button_group(GTK_RADIO_BUTTON(page_choice.period_radio)),
	    _("by month"));
  gtk_box_pack_start_defaults(GTK_BOX(vbox), page_choice.month_radio); 

  /* show the page and connect to control-flow managing functions */
  gtk_widget_show_all(page_choice.page);
  gnome_druid_append_page(GNOME_DRUID(druid.druid), 
			  GNOME_DRUID_PAGE(page_choice.page));
  gtk_signal_connect(GTK_OBJECT(page_choice.page), "cancel", 
		     GTK_SIGNAL_FUNC(report_druid_hide), NULL);
}

/* 
 * MAIN BUILD-FUNCTIONS **********************************************
 */

static void report_druid_build_window(void)
{
  druid.druid = gnome_druid_new();
  gtk_widget_show(druid.druid);

  druid.window = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_window_set_title(GTK_WINDOW(druid.window), _("Report Druid"));
  gtk_container_add(GTK_CONTAINER(druid.window), druid.druid);
}

static void report_druid_build(void)
{
  if (druid_already_built) 
    return;

  /* build the pages */
  report_druid_build_window();
  report_druid_build_start();
  report_druid_build_choice();
  report_druid_build_period_df();
  report_druid_build_finish();

  /* set flow order */
  report_druid_set_next(GNOME_DRUID_PAGE(page_start.page), 
			GNOME_DRUID_PAGE(page_choice.page));
  report_druid_set_back(GNOME_DRUID_PAGE(page_choice.page), 
			GNOME_DRUID_PAGE(page_start.page));
  gtk_signal_connect(GTK_OBJECT(page_choice.page), "next", 
		     GTK_SIGNAL_FUNC(report_druid_choice_fork), NULL);
  report_druid_set_back(GNOME_DRUID_PAGE(page_finish.page), 
			GNOME_DRUID_PAGE(page_choice.page));
  report_druid_set_next(GNOME_DRUID_PAGE(page_period_df.page), 
			GNOME_DRUID_PAGE(page_finish.page));
  report_druid_set_back(GNOME_DRUID_PAGE(page_period_df.page), 
			GNOME_DRUID_PAGE(page_choice.page));

  gnome_druid_set_page(GNOME_DRUID(druid.druid), 
		       GNOME_DRUID_PAGE(page_start.page));

  druid_already_built = TRUE;
}



/* 
 * FUNCTIONS ALTERING REPORT_GRP_LIST  ************************************
 */

static void report_druid_finish_period(void)
{
  GDate start, end;
  GDate period_start, period_end;
  gint groups_no;
  gfloat length;
  gint incl_label;
  gchar *label_fmt;
  gint i;

  g_date_clear(&start, 1);
  g_date_clear(&end, 1);
  g_date_clear(&period_start, 1);
  g_date_clear(&period_end, 1);

  groups_no = page_period_df.groups_no_adj->value;
  length = page_period_df.period_len_adj->value;
  date_entry_get_date(page_period_df.start_date, &start);
  date_entry_get_date(page_period_df.end_date, &end);
  incl_label = page_period_df.incl_label_adj->value;
  label_fmt = gtk_entry_get_text(GTK_ENTRY(page_period_df.label_entry));

  for (i = 0; i < groups_no; i++) {
    gchar start_text[50], end_text[50];
    ReportGroupData *rgd;
    rgd = g_new(ReportGroupData, 1);
    
    if (i == groups_no) {
      period_start = start;
      g_date_add_days(&period_start, 1);
    } else {
      period_start = end;
      if (GTK_TOGGLE_BUTTON(page_period_df.day_radio)->active) {
	g_date_subtract_days(&period_start, length * (i+1));
      } else {
	g_date_subtract_months(&period_start, floor(length*(i+1)));
	g_date_subtract_days(&period_start, 
			     floor(30 * (length*(i+1)-floor(length*(i+1)))));
      }
      g_date_add_days(&period_start, 1);
    }

    period_end = end;
    if (GTK_TOGGLE_BUTTON(page_period_df.day_radio)->active) {
      g_date_subtract_days(&period_end, length * i);
    } else {
      g_date_subtract_months(&period_end, floor(length * i));
      g_date_subtract_days(&period_end, 
			   floor(30 * (length*i-floor(length*i))));
    }

    misc_date_strftime(start_text, 49, &period_start);
    misc_date_strftime(end_text, 49, &period_end);
    snprintf(rgd->expr, PY_EXPRESSION_MAX_SIZE, 
	     "After(\"%s\") and Before(\"%s\")", 
	     start_text, end_text);

    if ((i % incl_label) == 0)
      g_date_strftime(rgd->label, REPORT_GROUP_LABEL_MAX_SIZE, 
		      label_fmt, &period_end);
    else
      strcpy(rgd->label, "");

    report_grp_axis_list = g_list_prepend(report_grp_axis_list, rgd);
  }
}

static void report_druid_finish_weekday(void)
{
  gint i;
  for (i = 0; i < 7; i++) {
    ReportGroupData *rgd;
    rgd = g_new(ReportGroupData, 1);
    
    if (i == 6)
      strcpy(rgd->label, nl_langinfo(ABDAY_1));
    else
      strcpy(rgd->label, nl_langinfo(ABDAY_2 + i));

    snprintf(rgd->expr, PY_EXPRESSION_MAX_SIZE, "Weekday() == %d", i+1);

    report_grp_axis_list = g_list_prepend(report_grp_axis_list, rgd);
  }
  report_grp_axis_list = g_list_reverse(report_grp_axis_list);
}

static void report_druid_finish_month(void)
{
  gint i;
  for (i = 0; i < 12; i++) {
    ReportGroupData *rgd;
    rgd = g_new(ReportGroupData, 1);
    
    strcpy(rgd->label, nl_langinfo(ABMON_1+i));

    snprintf(rgd->expr, PY_EXPRESSION_MAX_SIZE, "Month() == %d", i+1);

    report_grp_axis_list = g_list_prepend(report_grp_axis_list, rgd);
  }
  report_grp_axis_list = g_list_reverse(report_grp_axis_list);
}

static gboolean report_druid_finish(GtkWidget *widget, 
				    GtkWidget *event, 
				    gpointer callback_data)
{
  report_grp_axis_done();
  report_grp_axis_list = NULL;

  if (GTK_TOGGLE_BUTTON(page_choice.period_radio)->active)
    report_druid_finish_period();
  else if (GTK_TOGGLE_BUTTON(page_choice.weekday_radio)->active)
    report_druid_finish_weekday();
  else if (GTK_TOGGLE_BUTTON(page_choice.month_radio)->active)
    report_druid_finish_month();
  else
    g_assert_not_reached();

  g_assert(report_grp_axis_list != NULL);
  report_druid_hide(NULL, NULL, NULL);
  return TRUE;
}


/* 
 * INTERFACE TO THE DRUID *************************************************
 */

void report_druid_execute(void)
{
  report_druid_build();
  gtk_widget_show(druid.window);
}
